"""
This module contains the ``AdataDict`` class, which is basically a nested dictionary of anndata, with a 
few extra features to help restructure the nesting hierarchy and iterate over it.
"""

from functools import wraps

from anndata import AnnData

from .adata_dict_utils import to_nested_tuple, set_var_index_func, set_obs_index_func
from .adata_dict_fapply import adata_dict_fapply, adata_dict_fapply_return


class AdataDict(dict):
    """
    ``AdataDict`` is a dictionary-like container where values are ``AnnData`` objects. ``AdataDict`` inherits from :class:`dict`.

    This class provides three main functionalities:

    1. It has the ``set_hierarchy`` method to restructure the nesting hierarchy, and the ``hierarchy`` attribute to keep track.
    2. It behaves like an ``AnnData`` object by passing methods through to each ``AnnData`` in the dictionary.
    3. It has methods ``fapply(func, kwargs)`` and ``fapply_return(func, kwargs)`` that apply a given function ``func`` with arguments ``kwargs`` to each ``AnnData`` object in the :class:`AdataDict`.

    Parameters
    -----------
    data
        Dictionary with keys as tuples of indices.

    hierarchy
        Tuple or list indicating the order of indices in the keys of ``data``.

    """
    # See Also
    # --------

    # :func:`adata_dict_fapply` : The function underneath ``fapply`` that can be used separatley.
    # :func:`adata_dict_fapply_return` : The function underneath ``fapply_return`` that can be used separatley.

    def __init__(
    self,
    data: dict[tuple[int, ...], any] | None = None,
    hierarchy: tuple | list | None = None,
    ) -> None:
        """
        Initialize the ``AdataDict`` with data and hierarchy.

        Parameters
        ------------
        data
            Dictionary with keys as tuples of indices.

        hierarchy
            Tuple or list indicating the order of indices.

        Returns
        -------
        None
            Initializes the ``AdataDict`` object.
        """
        if data is None:
            data = {}
        super().__init__(data)  # Initialize the dict with data
        if hierarchy is not None:
            self._hierarchy = tuple(hierarchy)  # Tuple indicating the index hierarchy
        else:
            self._hierarchy = ()

    @property
    def hierarchy(self):
        """
        The hierarchy of the ``AdataDict``.

        This attribute is accessed as: ``adata_dict.hierarchy``.

        Returns
        --------
        The current hierarchy of the ``AdataDict`` as a tuple.

        Examples
        ---------

        .. code-block:: python

            adata_dict.hierarchy
            > ('donor', ('tissue'))
        """
        return self._hierarchy

    def _flatten(self, parent_key=()):
        flat_data = {}
        for key, value in self.items():
            full_key = parent_key + key
            if isinstance(value, AdataDict):
                flat_data.update(value._flatten(parent_key=full_key)) # pylint: disable=protected-access
            else:
                flat_data[full_key] = value
        return flat_data

    def flatten_nesting_list(self,
    nesting_list: list | tuple
    ) -> list:
        """
        Flatten a nested list or tuple into a single list.

        Parameters
        -----------
        nesting_list
            Nested list or tuple of hierarchy levels.

        Returns
        ---------
        Flattened list of hierarchy elements.
        """
        hierarchy = []
        for item in nesting_list:
            if isinstance(item, (list, tuple)):
                hierarchy.extend(self.flatten_nesting_list(item))
            else:
                hierarchy.append(item)
        return hierarchy

    def get_levels(self, nesting_list, levels=None, depth=0):
        """
        Get the levels of hierarchy based on the nesting structure.

        :param nesting_list: Nested list indicating the new hierarchy structure.
        :param levels: List to store the levels.
        :param depth: Current depth in recursion.
        :return: List of levels with hierarchy elements.
        """
        if levels is None:
            levels = []
        if len(levels) <= depth:
            levels.append([])
        for item in nesting_list:
            if isinstance(item, list):
                self.get_levels(item, levels, depth + 1)
            else:
                levels[depth].append(item)
        return levels

    def set_hierarchy(self,
    nesting_list: list
    ):
        """
        Rearrange the hierarchy of :class:`AdataDict` based on the provided nesting structure.

        Parameters
        ------------
        nesting_list
            Nested list indicating the new hierarchy structure.

        Examples
        ---------
        **Case 1: Flat hierarchy**

        .. code-block:: python

            adata_dict.set_hierarchy(["Donor", "Tissue"])
            print(adata_dict)
            > {
            >     ("Donor1", "Tissue1"): adata1,
            >     ("Donor1", "Tissue2"): adata2,
            >     ("Donor2", "Tissue1"): adata3,
            > }

        **Case 2: Nested hierarchy**

        .. code-block:: python

            adata_dict.set_hierarchy(["Donor", ["Tissue"]])  # Note the nested list here
            print(adata_dict)
            > {
            >     ("Donor1",): {
            >         ("Tissue1",): adata1,
            >         ("Tissue2",): adata2,
            >     },
            >     ("Donor2",): {
            >         ("Tissue1",): adata3,
            >     },
            > }

        **Case 3: Complex nested hierarchy**

        .. code-block:: python

            adata_dict.set_hierarchy(["Donor", ["Tissue", "Cell Type"]])  # Note the nested list here
            print(adata_dict)
            > {
            >     ("Donor1",): {
            >         ("Tissue1", "CellType1"): adata1,
            >         ("Tissue1", "CellType2"): adata2,
            >         ("Tissue2", "CellType3"): adata3,
            >     },
            >     ("Donor2",): {
            >         ("Tissue1", "CellType1"): adata4,
            >     },
            > }
        """

        # Flatten the nested data
        flat_data = self._flatten()
        self.clear()
        self.update(flat_data)

        # Flatten and extract the current hierarchy
        self._hierarchy = tuple(self.flatten_nesting_list(self._hierarchy))

        # Flatten the new hierarchy
        new_hierarchy = self.flatten_nesting_list(nesting_list)

        # Get the levels of the nesting structure
        levels = self.get_levels(nesting_list)
        old_hierarchy = self._hierarchy

        # Function to recursively create nested AdataDicts
        def create_nested_adata_dict(current_level, key_indices, value, level_idx):
            if level_idx == len(levels):
                return value  # Base case: return the value (AnnData object)
            level = levels[level_idx]
            level_length = len(level)
            level_key = tuple(key_indices[:level_length])
            remaining_key_indices = key_indices[level_length:]
            if level_key not in current_level:
                # Remaining hierarchy for nested AdataDict
                remaining_hierarchy = levels[level_idx + 1 :] if level_idx + 1 < len(levels) else []
                current_level[level_key] = AdataDict(hierarchy=remaining_hierarchy)
            # Recurse into the next level
            nested_dict = current_level[level_key]
            nested_value = create_nested_adata_dict(nested_dict, remaining_key_indices, value, level_idx + 1)
            if level_idx == len(levels) - 1:
                # At the last level, set the value
                current_level[level_key] = nested_value
            return current_level

        # Start building the new nested AdataDict
        new_data = AdataDict(hierarchy=new_hierarchy)
        for key, value in flat_data.items():
            # Map old indices to their values
            index_map = dict(zip(old_hierarchy, key))
            # Arrange indices according to the new hierarchy
            new_key_indices = [index_map[h] for h in new_hierarchy]
            # Recursively build the nested structure
            create_nested_adata_dict(new_data, new_key_indices, value, 0)

        # Update the hierarchy and data
        self._hierarchy = to_nested_tuple(nesting_list)  # Update with the nested structure
        # Replace the existing data in self with new_data
        self.clear()
        self.update(new_data)

    def flatten(self):
        """
        Flatten the hierarchy of :class:`AdataDict`.

        Examples
        ---------

        **Case 1: Nested hierarchy**

        .. code-block:: python

            #start with a nested hierarchy
            adata_dict.hierarchy
            > ('donor', ('tissue'))
            print(adata_dict)
            > {
            >     ("Donor1",): {
            >         ("Tissue1",): adata1,
            >         ("Tissue2",): adata2,
            >     },
            >     ("Donor2",): {
            >         ("Tissue1",): adata3,
            >     },
            > }

            #flatten adata_dict
            adata_dict.flatten()
            addata_dict.hierarchy
            > ('donor', 'tissue') #now the hierarchy is flat
            print(adata_dict)
            > {
            >     ("Donor1", "Tissue1"): adata1,
            >     ("Donor1", "Tissue2"): adata2,
            >     ("Donor2", "Tissue1"): adata3,
            > }
        """
        # Flatten the adata_dict
        flat_hierarchy = self.flatten_nesting_list(self.hierarchy)
        self.set_hierarchy(flat_hierarchy)

    def __getitem__(self, key):
        # Simplify access by converting non-tuple keys to tuple
        if not isinstance(key, tuple):
            key = (key,)
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        # Simplify setting by converting non-tuple keys to tuple
        if not isinstance(key, tuple):
            key = (key,)
        super().__setitem__(key, value)

    # def __getattr__(self, attr):
    #     def method(*args, **kwargs):
    #         results = {}
    #         for key, adata in self.items():
    #             if isinstance(adata, AdataDict):
    #                 # Recurse into nested AdataDict
    #                 results[key] = getattr(adata, attr)(*args, **kwargs)
    #             else:
    #                 func = getattr(adata, attr)
    #                 results[key] = func(*args, **kwargs)
    #         return results
    #     return method

    def __getattr__(self, attr):
        # First check if this is a property of AnnData
        if hasattr(AnnData, attr):
            anndata_attr = getattr(AnnData, attr)
            if isinstance(anndata_attr, property):
                # Handle property access
                return self._handle_property_access(attr)
        
        # If not a property, handle as a method
        return self._handle_method_access(attr)
    
    def _handle_property_access(self, attr):
        """Handle property access by returning a dictionary of property values."""
        results = {}
        for key, adata in self.items():
            if isinstance(adata, AdataDict):
                # Recurse into nested AdataDict
                results[key] = getattr(adata, attr)
            else:
                # Get property value directly
                results[key] = getattr(adata, attr)
        return results
    
    def _handle_method_access(self, attr):
        """Handle method access by returning a wrapper function."""
        def method(*args, **kwargs):
            results = {}
            for key, adata in self.items():
                if isinstance(adata, AdataDict):
                    # Recurse into nested AdataDict
                    results[key] = getattr(adata, attr)(*args, **kwargs)
                else:
                    func = getattr(adata, attr)
                    results[key] = func(*args, **kwargs)
            return results
        return method

    @wraps(adata_dict_fapply)
    def fapply(self, func, *, use_multithreading=True, num_workers=None, max_retries=0, **kwargs_dicts):
        """Wrapper for adata_dict_fapply."""
        return adata_dict_fapply(
            self,
            func,
            use_multithreading=use_multithreading,
            num_workers=num_workers,
            max_retries=max_retries,
            **kwargs_dicts,
        )

    @wraps(adata_dict_fapply_return)
    def fapply_return(self, func, *, use_multithreading=True, num_workers=None, max_retries=0, return_as_adata_dict=False, **kwargs_dicts):
        """Wrapper for adata_dict_fapply_return."""
        return adata_dict_fapply_return(
            self,
            func,
            use_multithreading=use_multithreading,
            num_workers=num_workers,
            max_retries=max_retries,
            return_as_adata_dict=return_as_adata_dict,
            **kwargs_dicts,
        )

    @wraps(set_var_index_func)
    def set_var_index(self, cols: str | list[str]):
        """Wrapper for set_var_index_func."""
        return set_var_index_func(self, cols)

    @wraps(set_obs_index_func)
    def set_obs_index(self, cols: str | list[str]):
        """Wrapper for set_obs_index_func."""
        return set_obs_index_func(self, cols)



"""
This module contains functions that operate across multiple adata, to make a sure that each adata has a .obs column with a set of categories that is shared between all the adata.
"""

from anndict.adata_dict import AdataDict
from anndict.automated_label_management.clean_single_column.in_adata_obs import map_cell_type_labels_to_simplified_set


def ai_unify_labels(
    adata_dict: AdataDict,
    label_columns: dict[tuple[str, ...], str],
    new_label_column: str,
    simplification_level: str = "unified, typo-fixed",
) -> dict:
    """
    Unifies cell type labels across multiple AnnData objects 
    by mapping them to a simplified, unified set of labels.

    Parameters
    ------------
    adata_dict
        An :class:`AdataDict`.

    label_columns
        :class:`dict` where keys should be the same as the keys of ``adata_dict`` and 
        values are the column names in the corresponding ``adata.obs`` containing the 
        original labels.

    new_label_column
        Name of the new column to be created in each ``adata.obs`` for storing the unified labels.
    
    simplification_level
        Instructions on how to unify the labels.

    Returns
    ---------
    A mapping :class:`dict` where the keys are the original labels and 
    the values are the unified labels.

    Notes
    -------
    Modifies each ``adata`` in ``adata_dict`` in-place by 
    adding ``adata.obs[new_label_column]`` with the unified label mapping.
    """
    # todo: use adata_dict_fapply instead of loops
    def get_unique_labels_from_obs_column(adata, label_column):
        return adata.obs[label_column].unique().tolist()

    def apply_mapping_to_adata(adata, mapping_dict, original_column, new_column, adt_key=None): # pylint: disable=unused-argument
        adata.obs[new_column] = adata.obs[original_column].map(mapping_dict)

    # Step 1: Aggregate all labels
    # aggregated_labels = adata_dict_fapply_return(adata_dict, get_unique_labels_from_obs_column, )
    aggregated_labels = []
    for key in adata_dict:
        labels = get_unique_labels_from_obs_column(adata_dict[key], label_columns[key])
        aggregated_labels.extend(labels)
    unique_labels_list = list(set(aggregated_labels))

    # Step 2: Get the mapping dictionary
    mapping_dict = map_cell_type_labels_to_simplified_set(unique_labels_list, simplification_level=simplification_level)

    # Step 3: Apply the mapping to each anndata in adata_dict
    for key in adata_dict:
        apply_mapping_to_adata(adata_dict[key], mapping_dict, label_columns[key], new_label_column)

    return mapping_dict



# llm_providers.py
"""
Manages configurations for different LLM providers supported by LangChain, including model paths,
class names, and initialization methods.
"""
from dataclasses import dataclass
from .base_llm_initializer import LLMProviderConfig
from .default_llm_initializer import DefaultLLMInitializer
from .custom_llm_initalizers import (  # type: ignore
    BedrockLLMInitializer,
    AzureMLLLMInitializer,
    GoogleGenAILLMInitializer,
)


@dataclass
class LLMProviders:
    """Collection of LLM provider configurations"""

    @classmethod
    def get_providers(cls) -> dict[str, LLMProviderConfig]:
        """Returns a dictionary mapping provider names to their LLM configurations.

        Provider configurations include the LangChain chat model class name, module path,
        and initializer class used to instantiate the model.

        Returns:
            Dict[str, LLMProviderConfig]: Map of provider names (e.g. 'openai', 'anthropic')
                to their corresponding LLMProviderConfig objects containing:
                - class_name (str): Name of LangChain chat model class
                - module_path (str): Import path for the chat model module
                - init_class (Type): Class used to initialize the chat model
        """
        return {
            "openai": LLMProviderConfig(
                class_name="ChatOpenAI",
                module_path="langchain_openai.chat_models",
                init_class=DefaultLLMInitializer,
            ),
            "anthropic": LLMProviderConfig(
                class_name="ChatAnthropic",
                module_path="langchain_anthropic.chat_models",
                init_class=DefaultLLMInitializer,
            ),
            "bedrock": LLMProviderConfig(
                class_name="ChatBedrockConverse",
                module_path="langchain_aws.chat_models.bedrock_converse",
                init_class=BedrockLLMInitializer,
            ),
            "google": LLMProviderConfig(
                class_name="ChatGoogleGenerativeAI",
                module_path="langchain_google_genai.chat_models",
                init_class=GoogleGenAILLMInitializer,
            ),
            "azureml_endpoint": LLMProviderConfig(
                class_name="AzureMLChatOnlineEndpoint",
                module_path="langchain_community.chat_models.azureml_endpoint",
                init_class=AzureMLLLMInitializer,
            ),
            "azure_openai": LLMProviderConfig(
                class_name="AzureChatOpenAI",
                module_path="langchain_community.chat_models.azure_openai",
                init_class=DefaultLLMInitializer,
            ),
            "cohere": LLMProviderConfig(
                class_name="ChatCohere",
                module_path="langchain_community.chat_models.cohere",
                init_class=DefaultLLMInitializer,
            ),
            "huggingface": LLMProviderConfig(
                class_name="ChatHuggingFace",
                module_path="langchain_community.chat_models.huggingface",
                init_class=DefaultLLMInitializer,
            ),
            "vertexai": LLMProviderConfig(
                class_name="ChatVertexAI",
                module_path="langchain_community.chat_models.vertexai",
                init_class=DefaultLLMInitializer,
            ),
            "ollama": LLMProviderConfig(
                class_name="ChatOllama",
                module_path="langchain_community.chat_models.ollama",
                init_class=DefaultLLMInitializer,
            ),
        }



"""
This module provides provider-specific LLM initializers.

Classes:
    ``BedrockLLMInitializer``: AWS Bedrock initialization and configuration
    ``AzureMLLLMInitializer``: Azure ML endpoint setup and configuration 
    ``GoogleGenAILLMInitializer``: Google Generative AI initialization

"""
import os
from typing import Any
import boto3
from langchain_community.chat_models.azureml_endpoint import (
    AzureMLEndpointApiType,
    LlamaChatContentFormatter,
)
from .base_llm_initializer import BaseLLMInitializer


class BedrockLLMInitializer(BaseLLMInitializer):
    """Initialization logic for AWS Bedrock"""

    MODELS_WITHOUT_SYSTEM_MESSAGES = {
        "amazon.titan-text-express-v1",
        "amazon.titan-text-lite-v1",
        "ai21.j2-ultra-v1",
    }

    def initialize(
        self, constructor_args: dict[str, Any], **kwargs
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        region_name = os.environ.get("LLM_REGION_NAME")
        aws_access_key_id = os.environ.get("LLM_AWS_ACCESS_KEY_ID")
        aws_secret_access_key = os.environ.get("LLM_AWS_SECRET_ACCESS_KEY")
        model_id = os.environ.get("LLM_MODEL")

        if not region_name:
            raise ValueError(
                "Bedrock requires LLM_REGION_NAME to be set in environment variables."
            )
        if not aws_access_key_id:
            raise ValueError(
                "Bedrock requires LLM_AWS_ACCESS_KEY_ID to be set in environment variables."
            )
        if not aws_secret_access_key:
            raise ValueError(
                "Bedrock requires LLM_AWS_SECRET_ACCESS_KEY to be set in environment variables."
            )
        if not model_id:
            raise ValueError("model_id (LLM_MODEL) is required for BedrockChat")

        kwargs["supports_system_messages"] = (
            model_id not in self.MODELS_WITHOUT_SYSTEM_MESSAGES
        )

        session = boto3.Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
        )

        bedrock_client = session.client("bedrock-runtime")
        allowed_params = ["model", "client", "streaming", "callbacks"]
        filtered_args = {
            k: v for k, v in constructor_args.items() if k in allowed_params
        }
        filtered_args["client"] = bedrock_client

        rate_limiter = self.create_rate_limiter(constructor_args)
        filtered_args["rate_limiter"] = rate_limiter

        return filtered_args, kwargs


class AzureMLLLMInitializer(BaseLLMInitializer):
    """Initialization logic for Azure ML endpoints"""

    def initialize(
        self, constructor_args: dict[str, Any], **kwargs
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        endpoint_name = constructor_args.pop("endpoint_name", None)
        region = constructor_args.pop("region", None)
        api_key = constructor_args.pop("api_key", None)

        if not all([endpoint_name, region, api_key]):
            raise ValueError(
                "AzureML requires endpoint_name, region, and api_key to be set."
            )

        constructor_args["endpoint_url"] = (
            f"https://{endpoint_name}.{region}.inference.ai.azure.com/v1/chat/completions"
        )
        constructor_args["endpoint_api_type"] = AzureMLEndpointApiType.serverless
        constructor_args["endpoint_api_key"] = api_key
        constructor_args["content_formatter"] = LlamaChatContentFormatter()

        return constructor_args, kwargs


class GoogleGenAILLMInitializer(BaseLLMInitializer):
    """Initialization logic for Google Generative AI"""

    def initialize(
        self, constructor_args: dict[str, Any], **kwargs
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        if "max_tokens" in kwargs:
            constructor_args["max_output_tokens"] = kwargs.pop("max_tokens")
        if "temperature" in kwargs:
            constructor_args["temperature"] = kwargs.pop("temperature")

        kwargs["supports_system_messages"] = False
        os.environ["GRPC_VERBOSITY"] = "ERROR"

        rate_limiter = self.create_rate_limiter(constructor_args)
        constructor_args["rate_limiter"] = rate_limiter

        return constructor_args, {}



"""
This module contains functions that use LLMs to automate label management within and between dataframes.
All of these functions are for processing category labels in pd dfs. We separate these functions
into 3 categories based on their input, see the left navigation
sidebar for the docs, and a short description below.

The functions are all broadly based around easing friction when working with categorical labels.
In the context of single cell RNA sequencing data, some common examples include:

- "My cell type labels have typos."
    - i.e. ``T cells`` and ``T cels``
- "I downloaded data from 3 different groups, and they each used a different label for the same cell type"
    - i.e. ``macrophage``, ``Macrophage.``, ``MΦ``
- "I want to coarsen the cell type labels."
    - i.e. Map ``CD8+ T cells`` and ``CD4+ T cells`` to a single category called ``T cells``


Single Column
---------------
**Input:** a single column.
    - These functions may generate more than one output column/category mapping, but they only look at a single set of input categories/labels.

Within Adata
--------------
**Input:** multiple columns within a single :class:`DataFrame` (i.e. ``.obs`` from a single ``adata``)
    - These functions process multiple columns within a single ``DataFrame``, usually outputting a single column that resolves differences among the input columns.

Between Adata
---------------
**Input** multiple DataFrames (i.e. ``.obs`` from several ``adata``)
    - These functions process columns across :class:`DataFrames`, usually inserting a new column into each that shares labels/categories across the :class:`DataFrames`.
"""

#post-process single categorical columns
from .clean_single_column import (
    simplify_obs_column,
    simplify_var_index,
    create_label_hierarchy,
    map_cell_type_labels_to_simplified_set,
    map_gene_labels_to_simplified_set,

)

#make categorical columns within a df share a common set of labels
from .unify_columns_within_adata import (
    ensure_label_consistency_main,
    ensure_label_consistency_adata,

)

#make categorical columns between dfs share a common set of labels
from .unify_columns_between_adata import (
    ai_unify_labels
)

__all__ = [
    # from clean single column
    # --------------------------
    # in_adata_obs.py
    "simplify_obs_column",
    "create_label_hierarchy",
    "map_cell_type_labels_to_simplified_set",

    # in_adata_var.py
    "simplify_var_index",
    "map_gene_labels_to_simplified_set",

    # from unify_columns_within_adata.py
    "ensure_label_consistency_main",
    "ensure_label_consistency_adata",

    # from unify_columns_between_adata.py
    "ai_unify_labels",
]



"""
This module contains and adata_dict wrappers for :mod:`anndata`.
Some of these ease manipulation of and information retrieval from :class:`AnnData`.
"""

import re

import numpy as np
import pandas as pd

from anndata import AnnData

from anndict.utils.anndictionary_ import enforce_semantic_list


def remove_genes(
    adata: AnnData,
    genes_to_remove: list[str],
    adt_key: tuple[str,...] | None = None
) -> None:
    """
    Remove specified genes from an :class:`AnnData` object in-place.

    Parameters
    -----------
    adata
        The :class:`AnnData` object to modify.

    genes_to_remove
        A list of gene names to remove.

    adt_key
        Used by :func:`adata_dict_fapply` and :func:`adata_dict_fapply_return` when passing this function.

    Returns
    --------
    ``None``

    Notes
    ------
    Modifies ``adata`` in-place.
    """
    # Get the list of genes to remove that are actually in the dataset
    genes_to_remove = adata.var_names.intersection(genes_to_remove)

    # Remove the specified genes
    # (the only way to do this in-place for now is to use the protected member of the class)
    adata._inplace_subset_var(~adata.var_names.isin(genes_to_remove)) # pylint: disable=protected-access

    print(f"Removed {len(genes_to_remove)} genes from {adt_key}. {adata.n_vars} genes remaining.")


def add_col_to_adata_obs(
    adata: AnnData,
    indices: list[int] | tuple[int, ...],
    values: list[int | float | str | bool | pd.Timestamp] | tuple[int | float | str | bool | pd.Timestamp, ...],
    new_col_name: str
) -> None:
    """
    Adds a column to ``adata.obs`` for given indices.

    Parameters
    -----------
    adata
        An :class:`AnnData`.

    indices
        Array of indices where labels will be assigned.

    values
        Array of labels corresponding to the indices.

    new_col_name
        Name of the new column to be created in ``adata.obs``.


    Returns
    --------
    ``None``

    Notes
    ------
    Modifies ``adata`` in-place.

    See Also
    ---------
    :func:`add_col_to_adata_var` : to add a column to ``adata.var``.
    :func:`add_col_to_pd_df` : the underlying function that adds a column to a :class:`DataFrame`.
    """
    add_col_to_pd_df(adata.obs, indices, values, new_col_name)


def add_col_to_adata_var(
    adata: AnnData,
    indices: list[int | str] | tuple[int | str, ...],
    values: list[int | float | str | bool | pd.Timestamp] | tuple[int | float | str | bool | pd.Timestamp, ...],
    new_col_name: str
) -> None:
    """
    Adds a column to ``adata.var`` for given indices.

    Parameters
    -----------
    adata
        An :class:`AnnData`.

    indices
        Array of indices where labels will be assigned.

    values
        Array of labels corresponding to the indices.

    new_label_key
        Name of the column in adata.var where the labels will be stored.

    Returns
    --------
    ``None``

    Notes
    ------
    Modifies ``adata`` in-place.

    See Also
    ---------
    :func:`add_col_to_adata_obs` : to add a column to ``adata.obs``.
    :func:`add_col_to_pd_df` : the underlying function that adds a column to a :class:`DataFrame`.
    """
    add_col_to_pd_df(adata.var, indices, values, new_col_name)


def add_col_to_pd_df(
    pd_df: pd.DataFrame,
    indices: list[int | str] | tuple[int | str, ...],
    values: list[int | float | str | bool | pd.Timestamp] | tuple[int | float | str | bool | pd.Timestamp, ...],
    new_col_name: str
) -> None:
    """
    Adds a column to ``pd_df`` for given indices.

    Parameters
    -----------
    pd_df
        An :class:`DataFrame`.

    indices
        Array of indices where labels will be assigned.

    values
        Array of labels corresponding to the indices.

    new_col_name
        Name of the new column to be created in ``pd_df``.


    Returns
    --------
    ``None``

    Notes
    ------
    Modifies ``pd_df`` in-place.
    """
    dtype = pd.Series(values[0]).dtype

    # Choose appropriate NA value and dtype
    if np.issubdtype(dtype, np.integer):
        dtype = pd.Int64Dtype()
        fill_value = pd.NA
    elif np.issubdtype(dtype, np.floating):
        fill_value = np.nan
    elif np.issubdtype(dtype, np.bool_):
        dtype = pd.BooleanDtype()
        fill_value = pd.NA
    else:
        fill_value = pd.NA

    full_series = pd.Series(
        fill_value,
        index=pd_df.index,
        dtype=dtype
    )

    indices_array = np.array(indices)

    # Use iloc only for integer indices, loc for everything else (strings, bools)
    if np.issubdtype(indices_array.dtype, np.integer):
        full_series.iloc[indices] = values
    else:
        full_series.loc[indices] = values

    pd_df[new_col_name] = full_series


def convert_obs_col_to_category(
    adata: AnnData,
    cols: str | list[str]
) -> None:
    """
    Convert column(s) in ``adata.obs`` to ``category`` dtype.

    Parameters
    -----------
    adata
        An :class:`AnnData`.

    cols
        The column name(s) in ``adata.obs`` to convert.

    Returns
    --------
    ``None``

    Notes
    ------
    Modifies ``adata`` in-place.
    """
    if isinstance(cols, str):  # Allow single string input as well
        cols = [cols]

    missing_cols = [col for col in cols if col not in adata.obs.columns]
    if missing_cols:
        raise ValueError(f"Columns {missing_cols} not found in adata.obs")

    for col in cols:
        adata.obs[col] = adata.obs[col].astype('category')


def convert_obs_col_to_string(
    adata: AnnData,
    cols: str | list[str] 
) -> None:
    """
    Convert column(s) in ``adata.obs`` to ``str`` dtype.

    Parameters
    -----------
    adata
        An :class:`AnnData`.

    cols
        The column name(s) in ``adata.obs`` to convert.

    Returns
    --------
    ``None``

    Notes
    ------
    Modifies ``adata`` in-place.
    """
    if isinstance(cols, str):  # Allow single string input as well
        cols = [cols]

    missing_cols = [col for col in cols if col not in adata.obs.columns]
    if missing_cols:
        raise ValueError(f"Columns {missing_cols} not found in adata.obs")

    for col in cols:
        adata.obs[col] = adata.obs[col].astype(str)


def convert_obs_index_to_str(
    adata: AnnData
) -> None:
    """
    Converts the index of ``adata.obs`` to ``str``.

    Parameters
    -----------
    adata
        An :class:`AnnData`.

    Returns
    --------
    ``None``

    Notes
    ------
    Modifies ``adata`` in-place.
    """
    adata.obs.index = adata.obs.index.astype(str)


def get_adata_columns(
    adata: AnnData,
    starts_with: str | list[str] | None = None,
    ends_with: str | list[str] | None = None,
    contains: str | list[str] | None = None,
    not_starts_with: str | list[str] | None = None,
    not_ends_with: str | list[str] | None = None,
    not_contains: str | list[str] | None = None
) -> list[str]:
    """
    A simple string matching and filtering function to get column names from ``adata.obs``. 
    The ``not_`` parameters are used to exclude columns from the results **after** the 
    inclusion filters have already been applied.

    Parameters
    ----------
    adata
        An AnnData object.

    starts_with
        Include columns that start with this.

    ends_with
        Include columns that end with this.

    contains
        Include columns that contain this.

    not_starts_with
        Exclude columns that start with this.

    not_ends_with
        Exclude columns that end with this.

    not_contains
        Exclude columns that contain this.

    Returns
    -------
    A list of column names from ``adata.obs`` that match the specified criteria.

    Examples
    --------
    .. code-block:: python

        import anndict as adt

        adata.obs.columns
        > Index(['cell_type', 'total_counts', 'total_counts_mt', 'total_counts_protein_coding',  
             'pct_counts_mt', 'pct_counts_protein_coding'], dtype='object')

        #get all total counts columns
        get_adata_columns(adata, starts_with='total_counts')
        > ['total_counts', 'total_counts_mt', 'total_counts_protein_coding']

        #get all mitochondrial columns
        get_adata_columns(adata, contains='mt')
        > ['total_counts_mt', 'pct_counts_mt']

        #get all columns that are not total counts
        get_adata_columns(adata, contains = '', not_starts_with='total_counts') #contains = '' will match all columns, then from this list, we exclude columns that start with ``'total_counts'``
        > ['cell_type', 'pct_counts_mt', 'pct_counts_protein_coding']
    """
    columns = adata.obs.columns
    matched_columns = []

    # Convert string inputs to single-element lists
    starts_with = [starts_with] if isinstance(starts_with, str) else starts_with
    ends_with = [ends_with] if isinstance(ends_with, str) else ends_with
    contains = [contains] if isinstance(contains, str) else contains
    not_starts_with = [not_starts_with] if isinstance(not_starts_with, str) else not_starts_with
    not_ends_with = [not_ends_with] if isinstance(not_ends_with, str) else not_ends_with
    not_contains = [not_contains] if isinstance(not_contains, str) else not_contains

    if starts_with:
        for start in starts_with:
            matched_columns.extend([col for col in columns if col.startswith(start)])

    if ends_with:
        for end in ends_with:
            matched_columns.extend([col for col in columns if col.endswith(end)])

    if contains:
        for contain in contains:
            matched_columns.extend([col for col in columns if contain in col])

    if not_starts_with:
        for start in not_starts_with:
            matched_columns = [col for col in matched_columns if not col.startswith(start)]

    if not_ends_with:
        for end in not_ends_with:
            matched_columns = [col for col in matched_columns if not col.endswith(end)]

    if not_contains:
        for contain in not_contains:
            matched_columns = [col for col in matched_columns if contain not in col]

    return list(set(matched_columns))


def filter_gene_list(
    adata: AnnData,
    gene_list: list[str]
) -> list[str]:
    """
    Filter a list of genes based on their presence in ``adata.var.index``. Will also update list with suffixes if needed to match gene names in ``adata.var.index``

    Parameters
    -----------
        adata
            An :class:`AnnData`.

        gene_list
            List of gene names to be filtered and updated with possible unique suffixes.

    Returns
    --------
    Updated list of genes found in ``adata.var.index``, including suffix variations.
    """
    enforce_semantic_list(adata.var.index)
    updated_gene_list = []
    for gene in gene_list:
        # Create a regex pattern to match the gene name and its possible unique suffixes, case-insensitive
        pattern = re.compile(r'^' + re.escape(gene) + r'(-\d+)?$', re.IGNORECASE)

        # Find all matching genes in adata.var.index
        matching_genes = [g for g in adata.var.index if pattern.match(g)]

        if matching_genes:
            updated_gene_list.extend(matching_genes)
        # else:
            # print(f"Gene '{gene}' not found in adata.var.index after making unique.")

    # Remove any duplicates in the updated marker list
    updated_gene_list = list(set(updated_gene_list))
    return updated_gene_list
